import argparse
import ctypes
import math
import mmap
import os
import pickle
import sys
from array import array
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from os import cpu_count, makedirs
from pathlib import Path
from typing import Any, Final, Iterable, Literal, Optional, Union
from zipfile import ZipFile

# Threshold to choose mode
THRESHOLD: Final = 0.5

# Aliases
ReadOnlyBuffer = bytes
if sys.version_info >= (3, 8):
    WriteableBuffer = (
        Union[bytearray, memoryview, mmap.mmap,
              pickle.PickleBuffer]
    )
else:
    WriteableBuffer = Union[bytearray, memoryview, array.array[
        Any], mmap.mmap, ctypes._CData]
ReadableBuffer = Union[ReadOnlyBuffer, WriteableBuffer]


# Exceptions
class UnableToCountCores(BaseException):
    """Unable to automatically find number of cores."""
    pass


class IncorrectNumberOfCores(BaseException):
    """Entered number more than there are cores on computer."""
    pass


class EmptyArchiveError(BaseException):
    """Archive is empty. Nothing to unzip."""
    pass


class AllFilesAreEmpty(BaseException):
    """All files in archive is 0 bytes."""
    pass


class CannotUseZeroCores(BaseException):
    """Number of cores cannot be zero"""
    pass


class Unzipper:

    """Unzipper prototype."""

    def __init__(self, zip_archive: str, path: str,
                 threads: Optional[int]) -> None:
        self._path = path  # Where to unpack
        self._zip_archive = zip_archive  # Where zip archive is
        self._cpu = cpu_count()  # CPU count for automatic work
        # Check for all situations, raise errors is smth is wrong
        if threads is None and self._cpu is None:
            raise UnableToCountCores("Enter cores or threads implicitly.")
        elif threads is None and self._cpu is not None:
            self._threads = min(self._cpu + 4, 32)
        elif threads is not None:
            if threads == 0:
                raise CannotUseZeroCores("Enter appropriate number.")
            else:
                self._threads = threads

    # save file to disk
    def _save_file(self, data: ReadableBuffer, filename: str) -> None:
        # create a path
        filepath = os.path.join(self._path, filename)
        # write to disk
        with open(filepath, 'wb') as file:
            file.write(data)


class MultiThreadUnzipper(Unzipper):

    """Unzipper for high compression levels"""

    # unzip files from an archive
    def __unzip_files(self, handle: ZipFile, filenames: Iterable[str]) -> None:
        # unzip multiple files
        for filename in filenames:
            # unzip the file
            handle.extract(filename, self._path)

    # unzip a large number of files
    def unzip(self) -> None:
        # open the zip file
        with ZipFile(self._zip_archive, 'r') as handle:
            # list of all files to unzip
            files = handle.namelist()
            if len(files) == 0:
                raise EmptyArchiveError("Archive is empty.")
            # determine chunksize
            chunksize = math.ceil(len(files) / self._threads)
            # start the thread pool
            with ThreadPoolExecutor(self._threads) as exe:
                # split the copy operations into chunks
                for i in range(0, len(files), chunksize):
                    # select a chunk of filenames
                    filenames = files[i:(i + chunksize)]
                    # submit the batch copy task
                    _ = exe.submit(self.__unzip_files, handle, filenames)


class CombinedUnzipper(Unzipper):

    """Unzipper for low compression levels"""

    def __init__(self, zip_archive: str, path: str, processes: Optional[int],
                 threads: Optional[int]) -> None:
        # Takes Unzipper constructor
        super().__init__(path, zip_archive, threads)
        # Checking for all types of input, raises errors if smth is wrong
        if processes is None and self._cpu is None:
            raise UnableToCountCores("Enter cores or threads implicitly.")
        elif processes is None and self._cpu is not None:
            self.__processes = self._cpu
        elif processes is not None and self._cpu is None:
            if processes == 0:
                raise CannotUseZeroCores("Enter appropriate number.")
            else:
                self.__processes = processes
            print("Attention!!! Unable to count cores. Please ensure you \
                entered correct number, else it can lead to \
                undefined behaviour.")
        elif processes is not None and self._cpu is not None:
            if processes > self._cpu:
                raise IncorrectNumberOfCores("Entered more cores than there \
                    are on computer.")
            elif processes == 0:
                raise CannotUseZeroCores("Enter appropriate number.")
            else:
                self.__processes = processes

    # unzip files from an archive
    def __unzip_files(self, filenames: Iterable[str]) -> None:
        # open the zip file
        with ZipFile(self._zip_archive, 'r') as handle:
            # create a thread pool
            with ThreadPoolExecutor(self._threads) as exe:
                # unzip each file
                for filename in filenames:
                    # decompress data
                    data = handle.read(filename)
                    # save to disk
                    _ = exe.submit(self._save_file, data, filename)

    # unzip a large number of files
    def unzip(self) -> None:
        # create the target directory
        makedirs(self._path, exist_ok=True)
        # open the zip file
        with ZipFile(self._zip_archive, 'r') as handle:
            # list of all files to unzip
            files = handle.namelist()
            if len(files) == 0:
                raise EmptyArchiveError("Archive is empty.")
        # determine chunksize
        chunksize = math.ceil(len(files) / self.__processes)
        # start the thread pool
        with ProcessPoolExecutor(self.__processes) as exe:
            # split the copy operations into chunks
            for i in range(0, len(files), chunksize):
                # select a chunk of filenames
                filenames = files[i:(i + chunksize)]
                # submit the batch copy task
                _ = exe.submit(self.__unzip_files, filenames)


class Controller:

    """Decides what subprogram to use using precalculated compression level"""

    def __init__(self, mode: Literal["mt", "cmbd", None],
                 zip_archive: str) -> None:
        self.mode = mode  # Initial mode or None if not specified
        self.zip_archive = zip_archive  # Where zip archive is

    def __calculate_compression(self) -> float:

        """Analyzes zip archive returning compression level"""

        compression = []
        with ZipFile(self.zip_archive, "r") as handle:
            if len(handle.namelist()) == 0:
                # No files in archive
                raise EmptyArchiveError("Archive is empty.")
            for file in handle.namelist():
                info = handle.getinfo(file)
                try:
                    # In case file is 0 bytes don't append list
                    compression.append(info.compress_size / info.file_size)
                except ZeroDivisionError:
                    pass
        if len(compression) == 0:
            # In case ZIP archive is empty
            raise AllFilesAreEmpty("All files are 0 bytes.")
        return (sum(compression) / len(compression))

    def get_compression(self) -> float:

        """Public function working as interface"""

        return self.__calculate_compression()


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(prog="fast unzipper",
                                     description="unzips ZIP archives",
                                     )
    parser.add_argument('archive_path', metavar='path', type=Path,
                        help="path to ZIP archive")
    parser.add_argument('-p', '--n_proc', default=None, type=int, metavar="",
                        help='number of processes')
    parser.add_argument('-t', '--n_threads', default=None, type=int,
                        metavar="",
                        help='number of threads for tasks in one process')
    parser.add_argument('-d', '--outdir', default=Path('./ZIP_unpack'),
                        metavar="",
                        type=Path, help='output directory to put unpacked \
                            data')
    parser.add_argument('-m', '--mode', default=None, metavar="",
                        type=str, help='mode in which program should work:\
                            "mt" or "cmbd"')
    args = parser.parse_args()

    assert args.archive_path.exists(), f'{args.archive_path}'
    return args


def main() -> None:
    args = parse_args()
    if args.mode is None:
        # If mode isn't specified controller decides what to use
        compression = (Controller(args.mode, args.archive_path)
                       .get_compression())
        if compression > THRESHOLD:
            MultiThreadUnzipper(args.archive_path, args.outdir,
                                args.n_threads).unzip()
        elif compression <= THRESHOLD:
            CombinedUnzipper(args.archive_path, args.outdir, args.n_proc,
                             args.n_threads).unzip()
    # If mode is given no controller is needed
    elif args.mode == "mt":
        MultiThreadUnzipper(args.archive_path, args.outdir,
                            args.n_threads).unzip()
    elif args.mode == "cmbd":
        CombinedUnzipper(args.archive_path, args.outdir, args.n_proc,
                         args.n_threads).unzip()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['fast_unzip']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['fast_unzip = src.fast_unzip.fast_unzip:main']}

setup_kwargs = {
    'name': 'fast-unzip',
    'version': '1.2.0',
    'description': 'Fast multithread/multiprocess unzipper',
    'long_description': '# Fast unzipper\n\nFast unzipper is a Python CLI tool you could use to effectively and fast unzip ZIP archives.\n\n## When will it be helpful?\n\nIt proved to be useful working with huge amount of relatively small files for it distributes load among processes and threads to provide higher speeds than standard unzip.\n\nHowever, should you need to unzip archive with only few files, that tool isn\'t probably for you because it won\'t be possible to distribute 1 or 2 files in archive among threads.\n\nMoreover, changing threads(e.g. 8) when there are 8 files will lead to decrease in performance rather than profit in time.\n\nUse it when it\'s possible to adequately split your files.\n\n\n## Installation\n\nUse the package manager [pip](https://pip.pypa.io/en/stable/) to install fast_unzip.\n\n```bash\npip install fast_unzip\n```\n\n## Usage\nIr\'s compulsory you specify the archive PATH like this.\n\nMacOS/Linux\n```bash\npython3 -m fast_unzip test.zip\n```\nWindows\n```bash\npython -m fast_unzip test.zip\n```\nBeing started this way it will use standard mode which means it will decide automatically which mode to use depending on compression level of your archive.\n\nStandard directory for unpacking is ./ZIP_unpack.\nYou can specify it with -d flag.\n```bash\npython -m fast_unzip test.zip -d ./../test\n```\n\nNevertheless, you can specify number of processes and threads you want to start using -p and -t flags.\n```bash\npython3 -m fast_unzip test.zip -p 4 -t 10\n```\nAlso, you can specify mode you want this tool to work. Maybe you know beforehand that compression level is low. You do it with -m flag.\n```bash\npython -m fast_unzip test.zip -m "mt"\n```\n## Recommendations\nThough, it\'s possible to choose mode, number of threads and number of processes manually, it\'s highly unrecommended, because if will affect the performance.\n\nNumber of threads is chosen as `min(32, os.cpu_count())`.\n\nNumber of processes is chosen as `os.cpu_count()`.\n\nFor some reasons `os.cpu_count()` can fail to determine your system characteristics. This way you\'ll be given an error and you need to specify this arguments implicitly. I highly recommend you use formula from above.\n\nIf you try to enter more processes than `os.cpu_count()`\nfound you\'ll be given an error. You can either choose an appropriate number of processes or leave it to program to decide.\n\n**!!! If it\'s impossible for `os.cpu_count()` to work and you enter inappropriate number of processes it will lead to undefined behaviour.**\n\nIf you know that archive you want to unpack is compressed less than 50% you can use `-m "cmbd"`, else `-m "mt"`. It will disable part of program doing analysis and increase performance.\n\nThank you for using our tool!\n\n## License\n\n[MIT](https://choosealicense.com/licenses/mit/)\n',
    'author': 'Nikita Pshenko',
    'author_email': 'pshenko1999@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Inferno2899/fast_unzip.git',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
